do
	Timer = {}

	DRIVER_VERSION = "1.3.0"

	MATRIX_NUM_CHANNELS = 8		-- 8 paths for supporting up to 8 zones on the amplifiers
	MATRIX_NUM_SOURCES = 32

	MATRIX_INPUT_BINDING_FIRST_CH_1 = 3001
	MATRIX_INPUT_BINDING_LAST_CH_1  = MATRIX_INPUT_BINDING_FIRST_CH_1 + MATRIX_NUM_SOURCES - 1
	MATRIX_INPUT_BINDING_FIRST = MATRIX_INPUT_BINDING_FIRST_CH_1
	MATRIX_INPUT_BINDING_LAST  = MATRIX_INPUT_BINDING_LAST_CH_1 + (MATRIX_NUM_CHANNELS - 1) * 100

	MATRIX_OUTPUT_BINDING_FIRST_CH_1 = 4001
	MATRIX_OUTPUT_BINDING_LAST_CH_1  = MATRIX_OUTPUT_BINDING_FIRST_CH_1 + MATRIX_NUM_SOURCES - 1
	MATRIX_OUTPUT_BINDING_FIRST = MATRIX_OUTPUT_BINDING_FIRST_CH_1
	MATRIX_OUTPUT_BINDING_LAST  = MATRIX_OUTPUT_BINDING_LAST_CH_1 + (MATRIX_NUM_CHANNELS - 1) * 100

	AMP_INPUT_BINDING_FIRST = 3000
	AMP_INPUT_BINDING_LAST  = 3031
end

function OnDriverInit()
end

function OnDriverLateInit()
	C4:UpdateProperty ("Driver Version", DRIVER_VERSION)

	KillAllTimers ()
	if (C4.AllowExecute) then C4:AllowExecute (true) end

	DEVICEID = C4:GetProxyDevices()
	if (not PersistData) then PersistData = {} end

	BuildBindings(false)

	CHANNELS = {}
	for i = 0, MATRIX_NUM_CHANNELS-1 do
		table.insert (CHANNELS, i, {})
	end

	-- We want the AV switch proxy to ask us which paths are valid. This ensures that paths from the inputs to the
	-- outputs use the same channel identifiers, which makes it possible to identify which room is making the source
	-- selection. Without the channels, the room on the output side is ambiguous and it would be much harder to
	-- reliably determine the correct zone identifier for the distributed source selection.
	C4:SendToProxy(5001, 'PROTOCOL_WILL_HANDLE_AV_VALID', {})
end

function OnDriverDestroyed()
end

function OnTimerExpired (idTimer)
	if (idTimer == Timer.BuildBindings) then
		BuildBindings(false)
	else
		CheckChannelAmpTimeouts (idTimer)
	end
end

-- Checks if the source device is connected to all amplifiers attached to the matrix outputs.
-- If the source device has multiple audio outputs this fails to differentiate between the case where
-- all the outputs are connected to all amplifiers and the case when all amplifiers are connected but not
-- via all of the source device's outputs. This is because the API lacks anything for getting or comparing
-- the bindings on both sides of a connection.
function is_source_global (source_device)
	-- Go through the output bindings on channel 1 only. The other channels always have the same bindings.
	for binding = MATRIX_OUTPUT_BINDING_FIRST_CH_1, MATRIX_OUTPUT_BINDING_LAST_CH_1 do
		local amplifiers = C4:GetBoundConsumerDevices(DEVICEID, binding)
		for amplifier, _ in pairs(amplifiers or {}) do
			if (amplifier ~= 0) then
				-- Go through all the sources on the amplifier and check if source_device is amongst them.
				local found = false
				for n_src = AMP_INPUT_BINDING_FIRST, AMP_INPUT_BINDING_LAST do
					if (tonumber(C4:GetBoundProviderDevice(amplifier, n_src)) == source_device) then
						found = true
						break
					end
				end
				if (not found) then
					return false
				end
			end
		end
	end
	return true
end

-- Returns true if other connections from the source device to the source amplifier exist.
function exist_other_connections(source_amplifier, src_binding_start, source_dev)
	for n_src = src_binding_start, AMP_INPUT_BINDING_LAST do
		if tonumber(C4:GetBoundProviderDevice(source_amplifier, n_src)) == source_dev then
			return true
		end
	end
	return false
end

-- Returns an array that maps proxy IDs to the respective device IDs.
function get_proxy_device_map()
	local map = {}
	local devices = C4:GetDevices()	-- could add filtering options here
	for k,v in pairs(devices) do
		local proxies = v.proxies or {}
		for k2,_ in pairs(proxies) do
			map[k2] = k
		end
	end
	return map
end

-- Finds a variable by name in a structure returned by C4:GetDeviceVariables()
function find_var(vars, name)
	for _,v in pairs(vars) do
		if v.name == name then
			return v.value
		end
	end
end

function BuildBindings(rebuild)
	local proxy_device_map = get_proxy_device_map()
	local amp_vars = {}
	-- Get a list of connected device IDs
	local connected_devices = {}
	for binding = MATRIX_INPUT_BINDING_FIRST_CH_1, MATRIX_INPUT_BINDING_LAST_CH_1 do
		-- Input binding: get the provider device, which is usually an Axium amplifier.
		local source_dev = tonumber (C4:GetBoundProviderDevice (DEVICEID, binding))
		if (source_dev ~= 0) then
			connected_devices[source_dev] = binding
		end
	end
	for binding = MATRIX_OUTPUT_BINDING_FIRST_CH_1, MATRIX_OUTPUT_BINDING_LAST_CH_1 do
		-- Output bindings: Get the consumer devices on each output (should only be one Axium amplifier).
		local consumer_devs = C4:GetBoundConsumerDevices (DEVICEID, binding)
		for dev, _ in pairs (consumer_devs or {}) do
			local dest_dev = tonumber(dev)
			if (dest_dev ~= 0) then
				connected_devices[dest_dev] = binding
			end
		end
	end

	local distributed_sources = {}
	local count = 0
	local source_amps = {}
	local source_counts = {}
	-- Go through the input bindings on channel 1 only. The other channels always have the same bindings.
	for binding = MATRIX_INPUT_BINDING_FIRST_CH_1, MATRIX_INPUT_BINDING_LAST_CH_1 do
		local source_amplifier = tonumber(C4:GetBoundProviderDevice(DEVICEID, binding))
		if (source_amplifier ~= 0) then
			local amp_dev_id = proxy_device_map[source_amplifier] or 0
			if amp_vars[amp_dev_id] == nil then
				amp_vars[amp_dev_id] = C4:GetDeviceVariables(amp_dev_id)
			end
			for n_src = AMP_INPUT_BINDING_FIRST, AMP_INPUT_BINDING_LAST do
				-- Get the source device ID, make sure it is valid and not already added to another amplifier. Also check that it
				-- isn't connected directly to all of the amplifiers (connected to the outputs of the matrix), otherwise the direct
				-- connections will always be used and there is no need for a distributed source definition. Note that a source device
				-- may be connected to multiple inputs on an amplifier since the source device may have multiple outputs, which
				-- potentially have different audio streams on. Also note that a source device will usually represent a single source
				-- but it may be multiple sources. Unfortunately the API does not provide a way to get the source device AND its
				-- output binding that is connected to a particular consumer and input binding.
				-- Also make sure the source is enabled for distribution.
				local source_dev = tonumber(C4:GetBoundProviderDevice(source_amplifier, n_src))
				if (source_dev ~= 0) then
					local ds_enable = find_var(amp_vars[amp_dev_id], string.format("BINDING_%u_DS_ENABLE", n_src)) or "1"
					if (source_amps[source_dev] or source_amplifier) == source_amplifier and ds_enable ~= "0" and not is_source_global(source_dev) then
						source_counts[source_dev] = (source_counts[source_dev] or 0) + 1
						source_amps[source_dev] = source_amplifier	-- don't add this device again if it is connected to another amp
						if (count == 32) then
							print("Too many distributed sources")
							break
						end
						if (distributed_sources[count] == nil) then
							distributed_sources[count] = {}
						end
						local def = distributed_sources[count]
						count = count + 1
						def.guid = string.format("%08X83B04B53999D4891B9A0AE8B", source_amplifier)
						def.source_input = n_src
						def.source_dev = source_dev
						def.amp_dev = source_amplifier
						def.name = C4:GetDeviceDisplayName(source_dev)
						def.matrix_input = binding
						if (source_counts[source_dev] > 1) or exist_other_connections(source_amplifier, n_src + 1, source_dev) then
							def.name = def.name .. " [" .. source_counts[source_dev] .. "]"
						end
					end
				end
			end
		end
		if (count == 32) then
			break
		end
	end
	local to_update = {}
	if (PersistData.distributed_sources ~= nil) and not rebuild then
		-- The table has already been generated previously so look for the things that have changed (if anything) and
		-- write only the changes. First look for items that were generated previously that no longer exist.
		local to_remove = {}
		for id, prev_source in pairs(PersistData.distributed_sources) do
			to_remove[id] = prev_source.amp_dev
			for _, source in pairs(distributed_sources) do
				if (prev_source.guid == source.guid and prev_source.source_input == source.source_input) then
					-- Found the source so don't remove it (remove it from the "remove" list)
					to_remove[id] = nil
					break
				end
			end
		end
		for id, amp_dev in pairs(to_remove) do
			PersistData.distributed_sources[id] = nil
			to_update[id] = amp_dev
		end
		-- Now look for items that now exist that did not exist previously or for ones that have been changed.
		local to_add = {}
		for temp_id, source in pairs(distributed_sources) do
			to_add[temp_id] = source
			for id, prev_source in pairs(PersistData.distributed_sources) do
				if (prev_source.guid == source.guid and prev_source.source_input == source.source_input) then
					-- Found the source so don't add it (remove it from the "add" list).
					to_add[temp_id] = nil
					-- If the name has changed then add to the "update" list.
					if (prev_source.name ~= source.name) then
						prev_source.name = source.name
						to_update[id] = source.amp_dev
					end
					if (prev_source.source_dev ~= source.source_dev) then
						prev_source.source_dev = source.source_dev
					end
					break
				end
			end
		end
		for _,source in pairs(to_add) do
			-- For each source in the add list, find an empty space in PersistData.distributed_sources
			for id = 0, 31 do
				if (PersistData.distributed_sources[id] == nil) then
					PersistData.distributed_sources[id] = source
					to_update[id] = source.amp_dev
					break
				end
			end
		end
	else
		-- The table has not been generated before or it is being rebuilt: write distributed_sources into persistent storage and
		-- send commands for all of the definitions to the amps. If it is being rebuilt then send commands for any previous definitions
		-- to ensure things get removed if fewer definitions will be created than before.
		if PersistData.distributed_sources ~= nil then
			-- Rebuilding: add the previous IDs to ensure anything not overwritten in the next step is deleted.
			for id, prev_source in pairs(PersistData.distributed_sources) do
				to_update[id] = prev_source.amp_dev
			end
		end
		PersistData.distributed_sources = distributed_sources
		for id, source in pairs(distributed_sources) do
			to_update[id] = source.amp_dev
		end
	end
	-- to_update contains the IDs of all the definitions that need to be written to the amps.
	for id, amp_dev in pairs(to_update) do
		local item = PersistData.distributed_sources[id] or {}
		local params = {}	-- copy the distributed source parameters along with the ID of the dist source
		params.id = id
		for k, v in pairs(item) do params[k] = v end
		-- Send the definition to the source amplifier driver. The driver will determine what the physical source ID is
		-- and broadcast out the definition to all amplifiers.
		C4:SendToDevice (amp_dev, "AXIUM_DISTRIBUTED_SOURCE_DEF", params, true)
		-- Send commands to all other amplifiers to flush out anything they might have for this entry in the table.
		for dev,_ in pairs(connected_devices) do
			if dev ~= amp_dev then
				local flush_params = {}
				flush_params.id = id
				C4:SendToDevice (dev, "FLUSH_DISTRIBUTED_SOURCE_DEF", flush_params, true)
			end
		end
	end
end

function get_destributed_source_id (source_device_id, source_input)
	if (source_device_id ~= nil and source_input ~= nil) then
		for id, source in pairs(PersistData.distributed_sources) do
			if (source.source_dev == source_device_id and source.source_input == source_input) then
				return id
			end
		end
	end
end

function AddTimer (timer, count, units, recur)
	local newTimer
	if (recur == nil) then recur = false end
	if (timer and timer ~= 0) then KillTimer (timer) end

	newTimer = C4:AddTimer (count, units, recur)
	return newTimer
end

function KillAllTimers ()
	for k,v in pairs (Timer or {}) do
		if (type (v) == 'number') then
			Timer [k] = KillTimer (Timer [k])
		end
	end
	--[[for _, thisQ in pairs (Qs or {}) do
		if (thisQ.ConnectingTimer and thisQ.ConnectingTimer ~= 0) then thisQ.ConnectingTimer = KillTimer (thisQ.ConnectingTimer) end
		if (thisQ.ConnectedTimer and thisQ.ConnectedTimer ~= 0) then thisQ.ConnectedTimer = KillTimer (thisQ.ConnectedTimer) end
	end]]
end

function KillTimer (timer)
	if (timer and type (timer) == 'number') then
		return (C4:KillTimer (timer))
	else
		return (0)
	end
end

-- Returns the channel ID from the input or output binding which is in the form 3CNN
-- or 4CNN where C is the channel.
function get_channel (idBinding)
	if (idBinding >= MATRIX_INPUT_BINDING_FIRST and idBinding <= MATRIX_OUTPUT_BINDING_LAST) then
		return math.floor(idBinding % 1000 / 100)
	end
end

function get_channel_source_amplifier_info (channel, source_amplifier)
	if (CHANNELS[channel][source_amplifier] == nil) then
		CHANNELS[channel][source_amplifier] = {}
		CHANNELS[channel][source_amplifier].timer = C4:AddTimer(1, "SECOND")
		print(string.format("Source selection on channel %d, amp ID %d", channel, source_amplifier))
	end
	return CHANNELS[channel][source_amplifier]
end

-- If all the required data are available, find the distributed source ID and send a message to the
-- destination amp to tell it to select it.
function process_channel_source_amplifier_info (channel, source_amplifier, info)
	if (info.source_device ~= nil and info.output_id ~= nil and PersistData.distributed_sources ~= nil) then
		local destributed_source_id = get_destributed_source_id (info.source_device, info.source_input)
		if (destributed_source_id ~= nil and destributed_source_id >= 0 and destributed_source_id < 32) then
			print(string.format("Selecting distributed source ID %d set on channel %d, amp ID %d", destributed_source_id, channel, source_amplifier))
			local params = {}
			params["OUTPUT_CHANNEL"] = channel
			params["SOURCE"] = destributed_source_id + 0x20		-- sources 0x20 .. 0x3F are distributed sources
			local dest_amplifiers = C4:GetBoundConsumerDevices (DEVICEID, info.output_id)
			for dest_amplifier, _ in pairs (dest_amplifiers or {}) do
				C4:SendToDevice (dest_amplifier, "AXIUM_SELECT_SOURCE", params, true)
			end
			C4:KillTimer (info.timer)
			CHANNELS[channel][source_amplifier] = nil
		end
	end
end

function CheckChannelAmpTimeouts (idTimer)
	for channel,amp_info in pairs(CHANNELS) do
		for source_amplifier,info in pairs(amp_info) do
			if (info.timer == idTimer) then
				print(string.format("Source selection timeout on channel %d, amp ID %d", channel, source_amplifier))
				CHANNELS[channel][source_amplifier] = nil
				return
			end
		end
	end
end

function ReceivedFromProxy(idBinding, sCommand, tParams)
	if (sCommand == 'BINDING_CHANGE_ACTION') then
		Timer.BuildBindings = AddTimer (Timer.BuildBindings, 1, 'SECONDS')
	elseif (sCommand == 'SET_INPUT') then
		-- The SET_INPUT commands from the AV switch proxy are not useful by themselves since it only connects local
		-- inputs to local outputs. This information must be used in combination with another message that is expected
		-- from the amplifier on the input side. The sequence of the messages is indeterminate.
		-- The input ID will be 3CNN and output will be 4CNN where C is the channel
		local input_id = tonumber(tParams["INPUT"] or 0)
		local output_id = tonumber(tParams["OUTPUT"] or 0)
		if (input_id ~= 0 and output_id ~= 0) then
			local channel = get_channel(output_id)
			if (channel == get_channel(input_id)) then
				local source_amplifier = C4:GetBoundProviderDevice(DEVICEID, input_id) or 0
				if (source_amplifier ~= 0) then
					local info = get_channel_source_amplifier_info(channel, source_amplifier)
					info.output_id = output_id
					print(string.format("Output ID %d set on channel %d, amp ID %d", output_id, channel, source_amplifier))
					process_channel_source_amplifier_info (channel, source_amplifier, info)
				end
			end
		end
	elseif (sCommand == "AXIUM_MATRIX_SOURCE") then
		-- Received source selection information from an amplifier on the input side of the matrix. This information
		-- is combined with information from SET_INPUT to determine what distributed source has been selected.
		local source_device = tonumber(tParams["SOURCE_DEV"] or 0)
		local source_amplifier = tonumber(tParams["AMP_DEV"] or 0)
		local source_input = tonumber(tParams["AMP_INPUT"] or 0)
		local channel = tonumber(tParams["CHANNEL"] or 0)
		if (source_device ~= 0 and source_amplifier ~= 0) then
			local info = get_channel_source_amplifier_info(channel, source_amplifier)
			info.source_device = source_device
			info.source_input = source_input
			print(string.format("Source device %d set on channel %d, amp ID %d, input %d", source_device, channel, source_amplifier, source_input))
			process_channel_source_amplifier_info (channel, source_amplifier, info)
		end
	elseif (sCommand == 'IS_AV_OUTPUT_TO_INPUT_VALID') then
		-- Inputs and outputs are in groups of 8 channels that are needed to represent the 8 zones of each
		-- amplifier attached to the output of the matrix. The inputs for one channel are not supposed to
		-- be connected to the outputs of another channel so return "True" or "False" for the various
		-- combinations that the proxy asks us about.
		local source_id = tonumber(tParams["Params_idSource"] or 0)
		local input_binding = tonumber(tParams["Consumer_idBinding"] or 0)
		local output_binding = tonumber(tParams["Provider_idBinding"] or 0)
		-- Make sure input and output bindings are in the expected magic number ranges
		if (input_binding >= MATRIX_INPUT_BINDING_FIRST and input_binding <= MATRIX_INPUT_BINDING_LAST
				and output_binding >= MATRIX_OUTPUT_BINDING_FIRST and output_binding <= MATRIX_OUTPUT_BINDING_LAST)
		then
			-- Check if the channel of the input and output bindings match. The IDs are 3CNN and 4CNN where C is the channel
			-- and NN is the audio source node.
			local channel = get_channel(input_binding)
			if (channel == get_channel(output_binding)) then
				-- Finally make sure input_binding and output_binding are connected to different devices since we want to
				-- avoid connecting a device to itself.
				local input_device = C4:GetBoundProviderDevice (DEVICEID, input_binding)
				local output_devices = C4:GetBoundConsumerDevices (DEVICEID, output_binding)
				if (input_device ~= nil and output_devices ~= nil) then
					-- Make sure the input device is not found amongst the output devices
					if (output_devices[input_device] == nil) then
						-- Now check if the path to source_id exists in the distributed source definitions via the specified matrix
						-- input binding. If the source is connected to multiple amplifiers and only one of them hosts a distributed
						-- source then we need to ensure that the path to the amplifier with the distributed source is used. If this
						-- is not the case then process_channel_source_amplifier_info will fail to find the distributed source ID to
						-- be selected.
						if source_id ~= 0 then
							for _,source in pairs(PersistData.distributed_sources) do
								if (source.source_dev == source_id) and source.matrix_input == input_binding then
									return "True"	-- This path is OK
								end
							end
						else
							-- This case is allowed in case Params_idSource is just set to 0.
							return "True"	-- This path is (hopefully) OK
						end
					end
				end
			end
		end
		return "False"		-- Do not use this path
	end
end

function ExecuteCommand(sCommand, tParams)
	if (sCommand == "LUA_ACTION") then
		if (tParams["ACTION"] == "ACTION_RebuildDistributedSources") then
			BuildBindings(true)
		end
	end
end
